/*
 * Copyright 2009-2009 the Fess Project and the Others.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package jp.sf.fess.task;

import java.io.File;
import java.io.Serializable;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;

import javax.annotation.Resource;

import jp.sf.fess.Constants;
import jp.sf.fess.db.exentity.CrawlingSession;
import jp.sf.fess.helper.CrawlingSessionHelper;
import jp.sf.fess.helper.FileSystemIndexHelper;
import jp.sf.fess.helper.OverlappingHostHelper;
import jp.sf.fess.helper.PathMappingHelper;
import jp.sf.fess.helper.SystemHelper;
import jp.sf.fess.helper.WebIndexHelper;
import jp.sf.fess.service.CrawlingSessionService;
import jp.sf.fess.service.PathMappingService;
import jp.sf.fess.solr.SolrServerGroup;
import jp.sf.fess.solr.SolrServerManager;
import jp.sf.fess.util.FessProperties;

import org.apache.commons.lang.time.DateUtils;
import org.seasar.chronos.core.TaskTrigger;
import org.seasar.chronos.core.annotation.task.Task;
import org.seasar.chronos.core.trigger.CCronTrigger;
import org.seasar.framework.container.SingletonS2Container;
import org.seasar.framework.util.StringUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

@Task
public class CrawlTask implements Serializable {

    private static final long serialVersionUID = 1L;

    private static final Logger logger = LoggerFactory
            .getLogger(CrawlTask.class);

    @Resource
    protected FessProperties crawlerProperties;

    @Resource
    protected WebIndexHelper webIndexHelper;

    @Resource
    protected FileSystemIndexHelper fileSystemIndexHelper;

    @Resource
    protected SolrServerManager solrServerManager;

    @Resource
    protected CrawlingSessionService crawlingSessionService;

    @Resource
    protected PathMappingService pathMappingService;

    private CCronTrigger trigger;

    public TaskTrigger getTrigger() {
        if (trigger == null) {
            trigger = new CCronTrigger(crawlerProperties.getProperty(
                    Constants.CRON_EXPRESSION_PROPERTY,
                    Constants.DEFAULT_CRON_EXPRESSION));
        } else {
            trigger.setExpression(crawlerProperties.getProperty(
                    Constants.CRON_EXPRESSION_PROPERTY,
                    Constants.DEFAULT_CRON_EXPRESSION));
        }
        return trigger;
    }

    public void setExpression(String cronExpression) {
        if (trigger != null)
            trigger.setExpression(cronExpression);
    }

    public String getExpression() {
        if (trigger != null) {
            trigger.getCronExpression();
        }
        return null;
    }

    public void doExecute() {
        SystemHelper systemHelper = SingletonS2Container
                .getComponent("systemHelper");
        if (systemHelper.readyCrawlProcess()) {
            Map<String, String> infoMap = null;
            // create session id
            SimpleDateFormat sdf = new SimpleDateFormat("yyyyMMddHHmmssSSS");
            String sessionId = sdf.format(new Date());
            systemHelper.setSessionId(sessionId);
            // store crawling session
            CrawlingSession crawlingSession = new CrawlingSession(sessionId);
            crawlingSessionService.store(crawlingSession);
            try {
                if (Constants.TRUE.equals(crawlerProperties.getProperty(
                        Constants.SNAPSHOT_REPLICATION_PROPERTY,
                        Constants.FALSE))) {
                    // replication
                    infoMap = doReplication(sessionId);
                } else {
                    infoMap = doCrawl(sessionId);
                }
            } finally {
                systemHelper.finishCrawlProcess();
                if (infoMap != null && !infoMap.isEmpty()) {
                    try {
                        CrawlingSessionHelper crawlingSessionHelper = SingletonS2Container
                                .getComponent("crawlingSessionHelper");
                        crawlingSessionHelper.put(sessionId, infoMap);
                    } catch (Exception e) {
                        logger.warn("Failed to store crawling information.", e);
                    }
                }
            }
        } else {
            logger.warn("Crawler is running now.");
        }
    }

    public Map<String, String> doReplication(String sessionId) {
        String snapshotPath = crawlerProperties.getProperty(
                Constants.SNAPSHOT_PATH_PROPERTY, null);
        if (StringUtil.isBlank(snapshotPath)) {
            if (logger.isWarnEnabled()) {
                logger.warn("A snapshot path is a blank.");
            }
            return null;
        }
        // TODO remote snapshot path
        SystemHelper systemHelper = SingletonS2Container
                .getComponent("systemHelper");
        File snapshotDir = systemHelper.getSnapshotDir(snapshotPath);
        if (!snapshotDir.exists()) {
            if (logger.isWarnEnabled()) {
                logger.warn(snapshotPath + " does not exist.");
            }
            return null;
        }

        if (logger.isInfoEnabled()) {
            logger.info("Starting Replication..");
        }

        Map<String, String> infoMap = new HashMap<String, String>();
        SimpleDateFormat dateFormat = new SimpleDateFormat(
                "yyyy-MM-dd'T'HH:mm:ss.SSSZ");
        long totalTime = System.currentTimeMillis();

        infoMap.put(Constants.REPLICATION_START_TIME, dateFormat
                .format(new Date()));
        boolean completed = false;
        try {
            // TODO search lock?

            SolrServerGroup solrServerGroup = solrServerManager
                    .getUpdateSolrServerGroup();

            // replication
            solrServerGroup.replicate(snapshotDir);

            // commit
            long startTime = System.currentTimeMillis();
            solrServerGroup.commit();
            startTime = System.currentTimeMillis() - startTime;
            if (logger.isInfoEnabled()) {
                logger.info("[EXEC TIME] replication commit time: " + startTime
                        + "ms");
            }

            String serverRotationStr = crawlerProperties.getProperty(
                    Constants.SERVER_ROTATION_PROPERTY, Constants.TRUE);
            if (Constants.TRUE.equalsIgnoreCase(serverRotationStr)) {
                // apply
                solrServerManager.applyNewServerGroup();
            }

            completed = true;
        } catch (Throwable t) {
            if (logger.isWarnEnabled()) {
                logger.warn("Interrupted a replication task.", t);
            }
        } finally {
            infoMap.put(Constants.REPLICATION_STATUS, completed ? Constants.T
                    : Constants.F);
            infoMap.put(Constants.REPLICATION_END_TIME, dateFormat
                    .format(new Date()));
            infoMap.put(Constants.REPLICATION_EXEC_TIME, Long.toString(System
                    .currentTimeMillis()
                    - totalTime));
        }
        if (logger.isInfoEnabled()) {
            logger.info("Finished Replication");
        }
        return infoMap;
    }

    public Map<String, String> doCrawl(String sessionId) {
        if (logger.isInfoEnabled()) {
            logger.info("Starting Crawler..");
        }

        PathMappingHelper pathMappingHelper = SingletonS2Container
                .getComponent("pathMappingHelper");

        SimpleDateFormat dateFormat = new SimpleDateFormat(
                "yyyy-MM-dd'T'HH:mm:ss.SSSZ");

        long totalTime = System.currentTimeMillis();
        Map<String, String> infoMap = new HashMap<String, String>();

        boolean completed = false;
        try {
            infoMap.put(Constants.CRAWLER_START_TIME, dateFormat
                    .format(new Date()));

            SolrServerGroup solrServerGroup = solrServerManager
                    .getUpdateSolrServerGroup();

            // setup path mapping
            pathMappingHelper.setPathMappingList(sessionId, pathMappingService
                    .getPathMappingList());

            // overlapping host
            try {
                OverlappingHostHelper overlappingHostHelper = SingletonS2Container
                        .getComponent("overlappingHostHelper");
                overlappingHostHelper.init();
            } catch (Exception e) {
                logger.warn("Could not initialize overlappingHostHelper.", e);
            }

            // crawl web
            infoMap.put(Constants.WEB_CRAWLER_START_TIME, dateFormat
                    .format(new Date()));
            webIndexHelper.crawl(sessionId, solrServerGroup);
            infoMap.put(Constants.WEB_CRAWLER_END_TIME, dateFormat
                    .format(new Date()));

            // crawl file system
            infoMap.put(Constants.FS_CRAWLER_START_TIME, dateFormat
                    .format(new Date()));
            fileSystemIndexHelper.crawl(sessionId, solrServerGroup);
            infoMap.put(Constants.FS_CRAWLER_END_TIME, dateFormat
                    .format(new Date()));

            // clean up
            String dayForCleanupStr = crawlerProperties.getProperty(
                    Constants.DAY_FOR_CLEANUP_PROPERTY, "1");
            int dayForCleanup = -1;
            try {
                dayForCleanup = Integer.parseInt(dayForCleanupStr);
            } catch (NumberFormatException e) {
            }
            if (dayForCleanup >= 0) {
                Date date = DateUtils.addDays(new Date(), -1 * dayForCleanup);
                String[] sessionIds = crawlingSessionService
                        .getSessionIdsBefore(date);
                SolrServerGroup selectSolrServerGroup = solrServerManager
                        .getSelectSolrServerGroup();
                for (String sid : sessionIds) {
                    // delete 
                    solrServerGroup.deleteByQuery("segment:" + sid);
                    if (selectSolrServerGroup != solrServerGroup) {
                        selectSolrServerGroup.deleteByQuery("segment:" + sid);
                    }
                }
            }

            // optimize
            String optimizeStr = crawlerProperties.getProperty(
                    Constants.OPTIMIZE_PROPERTY, Constants.TRUE);
            String commitStr = crawlerProperties.getProperty(
                    Constants.COMMIT_PROPERTY, Constants.TRUE);
            if (Constants.TRUE.equalsIgnoreCase(optimizeStr)) {
                infoMap.put(Constants.OPTIMIZE_START_TIME, dateFormat
                        .format(new Date()));
                long startTime = System.currentTimeMillis();
                solrServerGroup.optimize();
                startTime = System.currentTimeMillis() - startTime;
                infoMap.put(Constants.OPTIMIZE_END_TIME, dateFormat
                        .format(new Date()));
                infoMap.put(Constants.OPTIMIZE_EXEC_TIME, Long
                        .toString(startTime));
                if (logger.isInfoEnabled()) {
                    logger.info("[EXEC TIME] index optimize time: " + startTime
                            + "ms");
                }
            } else if (Constants.TRUE.equalsIgnoreCase(commitStr)) {
                infoMap.put(Constants.COMMIT_START_TIME, dateFormat
                        .format(new Date()));
                long startTime = System.currentTimeMillis();
                solrServerGroup.commit();
                startTime = System.currentTimeMillis() - startTime;
                infoMap.put(Constants.COMMIT_END_TIME, dateFormat
                        .format(new Date()));
                infoMap.put(Constants.COMMIT_EXEC_TIME, Long
                        .toString(startTime));
                if (logger.isInfoEnabled()) {
                    logger.info("[EXEC TIME] index commit time: " + startTime
                            + "ms");
                }
            } else {
                if (logger.isInfoEnabled()) {
                    logger.info("No index commit.");
                }
            }

            String serverRotationStr = crawlerProperties.getProperty(
                    Constants.SERVER_ROTATION_PROPERTY, Constants.TRUE);
            if (Constants.TRUE.equalsIgnoreCase(serverRotationStr)) {
                // apply
                solrServerManager.applyNewServerGroup();
            }

            if (logger.isInfoEnabled()) {
                logger.info("Finished Crawler");
            }
            completed = true;
        } catch (Throwable t) {
            if (logger.isWarnEnabled()) {
                logger.warn("Interrupted a crawl task.", t);
            }
        } finally {
            pathMappingHelper.removePathMappingList(sessionId);
            infoMap.put(Constants.CRAWLER_STATUS, completed ? Constants.T
                    : Constants.F);
            infoMap.put(Constants.CRAWLER_END_TIME, dateFormat
                    .format(new Date()));
            infoMap.put(Constants.CRAWLER_EXEC_TIME, Long.toString(System
                    .currentTimeMillis()
                    - totalTime));

        }
        return infoMap;
    }

    public void catchException(Exception e) {
        logger.error("Failed to execute crawl task.", e);
    }

}
